import time
from typing import Any, Callable, Dict, List, Optional

from django.core.cache import cache as default_cache
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest
from ninja.throttling import SimpleRateThrottle as BaseSimpleRateThrottle

from ninja_extra.conf import settings


class SimpleRateThrottle(BaseSimpleRateThrottle):
    """
    A simple cache implementation, that only requires `.get_cache_key()`
    to be overridden.

    The rate (requests / seconds) is set by a `rate` attribute on the Throttle
    class.  The attribute is a string of the form 'number_of_requests/period'.

    Period should be one of: ('s', 'sec', 'm', 'min', 'h', 'hour', 'd', 'day')

    Previous request information used for throttling is stored in the cache.
    """

    cache: Any = default_cache
    timer: Callable[[], float] = time.time
    cache_format: str = "throttle_%(scope)s_%(ident)s"
    scope: Optional[str] = None

    def __init__(self, rate: Optional[str] = None) -> None:
        if not rate and hasattr(self, "rate"):
            rate = getattr(self, "rate", None)
        super(SimpleRateThrottle, self).__init__(rate)

    def get_throttling_rates(self) -> Dict[str, Optional[str]]:
        rates = self.THROTTLE_RATES.copy()
        rates.update(settings.THROTTLE_RATES)

        return rates

    def get_cache_key(self, request: HttpRequest) -> Optional[str]:
        """
        Should return a unique cache-key which can be used for throttling.
        Must be overridden.

        May return `None` if the request should not be throttled.
        """
        raise NotImplementedError(".get_cache_key() must be overridden")

    def get_rate(self) -> Optional[str]:
        """
        Determine the string representation of the allowed request rate.
        """
        if not self.scope:
            msg = (
                "You must set either `.scope` or `.rate` for '%s' throttle"
                % self.__class__.__name__
            )
            raise ImproperlyConfigured(msg)

        _THROTTLE_RATES = self.get_throttling_rates()
        try:
            return _THROTTLE_RATES[self.scope]
        except KeyError as e:
            msg = "No default throttle rate set for '%s' scope" % self.scope
            raise ImproperlyConfigured(msg) from e

    def allow_request(self, request: HttpRequest) -> bool:
        """
        Implement the check to see if the request should be throttled.

        On success calls `throttle_success`.
        On failure calls `throttle_failure`.
        """
        if self.rate is None:
            return True

        return super().allow_request(request)

    def get_ident(self, request: HttpRequest) -> Optional[str]:
        """
        Identify the machine making the request by parsing HTTP_X_FORWARDED_FOR
        if present and number of proxies is > 0. If not use all of
        HTTP_X_FORWARDED_FOR if it is available, if not use REMOTE_ADDR.
        """

        xff = request.META.get("HTTP_X_FORWARDED_FOR")
        remote_addr = request.META.get("REMOTE_ADDR")
        num_proxies = settings.NUM_PROXIES

        if num_proxies is not None:
            if num_proxies == 0 or xff is None:
                return remote_addr
            addrs: List[str] = xff.split(",")
            client_addr = addrs[-min(num_proxies, len(addrs))]
            return client_addr.strip()

        return "".join(xff.split()) if xff else remote_addr


class AnonRateThrottle(SimpleRateThrottle):
    """
    Limits the rate of API calls that may be made by a anonymous users.

    The IP address of the request will be used as the unique cache key.
    """

    scope = "anon"

    def get_cache_key(self, request: HttpRequest) -> Optional[str]:
        if request.user and request.user.is_authenticated:
            return None  # Only throttle unauthenticated requests.

        return self.cache_format % {
            "scope": self.scope,
            "ident": self.get_ident(request),
        }


class UserRateThrottle(SimpleRateThrottle):
    """
    Limits the rate of API calls that may be made by a given user.

    The user id will be used as a unique cache key if the user is
    authenticated.  For anonymous requests, the IP address of the request will
    be used.
    """

    scope = "user"

    def get_cache_key(self, request: HttpRequest) -> Optional[str]:
        if request.user and request.user.is_authenticated:
            ident = request.user.pk
        else:
            ident = self.get_ident(request)

        return self.cache_format % {"scope": self.scope, "ident": ident}


class DynamicRateThrottle(SimpleRateThrottle):
    """
    Limits the rate of API calls by different amounts for various parts of
    the API. Set Throttle scope dynamically for an api endpoint
    """

    def __init__(self, rate: Optional[str] = None, scope: Optional[str] = None) -> None:
        self.scope = scope
        super().__init__(rate)

    def get_cache_key(self, request: HttpRequest) -> Optional[str]:
        """
        If `scope` is not set during initialization, don't apply this throttle.

        Otherwise, generate the unique cache key by concatenating the user id
        with the `.throttle_scope` property of the view.
        """
        if request.user and request.user.is_authenticated:
            ident = request.user.pk
        else:
            ident = self.get_ident(request)

        return self.cache_format % {"scope": self.scope, "ident": ident}



import threading
from typing import Optional

from django.conf import Settings, settings
from injector import Binder, Module, singleton

from ninja_extra.controllers.route.context import RouteContext

from .constants import ROUTE_CONTEXT_VAR

# @receiver([route_context_started, route_context_finished], sender=RouteContext)
# def route_context_handler(
#     *args: Any, route_context: Optional[RouteContext] = None, **kwargs: Any
# ) -> None:
#     app = cast(Any, apps.get_app_config("ninja_extra"))
#     app.ninja_extra_module.set_route_context(route_context)


class NinjaExtraModule(Module):
    def __init__(self) -> None:
        self._local = threading.local()

    def set_route_context(self, route_context: RouteContext) -> None:
        ROUTE_CONTEXT_VAR.set(route_context)

    def get_route_context(self) -> Optional[RouteContext]:
        return ROUTE_CONTEXT_VAR.get()

    def configure(self, binder: Binder) -> None:
        binder.bind(Settings, to=settings, scope=singleton)  # type: ignore
        binder.bind(RouteContext, to=lambda: self.get_route_context())  # type: ignore



from ninja.throttling import BaseThrottle

from .decorator import throttle
from .model import (
    AnonRateThrottle,
    DynamicRateThrottle,
    SimpleRateThrottle,
    UserRateThrottle,
)

__all__ = [
    "BaseThrottle",
    "DynamicRateThrottle",
    "UserRateThrottle",
    "SimpleRateThrottle",
    "AnonRateThrottle",
    "throttle",
]



import typing as t
from abc import ABC, abstractmethod

from django.db.models import Model as DjangoModel
from django.db.models import QuerySet
from pydantic import BaseModel as PydanticModel


class AsyncModelServiceBase(ABC):
    @abstractmethod
    async def get_one_async(self, pk: t.Any, **kwargs: t.Any) -> t.Any:
        pass

    @abstractmethod
    async def get_all_async(self, **kwargs: t.Any) -> t.Union[QuerySet, t.List[t.Any]]:
        pass

    @abstractmethod
    async def create_async(self, schema: PydanticModel, **kwargs: t.Any) -> t.Any:
        pass

    @abstractmethod
    async def update_async(
        self, instance: DjangoModel, schema: PydanticModel, **kwargs: t.Any
    ) -> t.Any:
        pass

    @abstractmethod
    async def patch_async(
        self, instance: DjangoModel, schema: PydanticModel, **kwargs: t.Any
    ) -> t.Any:
        pass

    @abstractmethod
    async def delete_async(self, instance: DjangoModel, **kwargs: t.Any) -> t.Any:
        pass


class ModelServiceBase(ABC):
    """
    Abstract service that handles Model Controller model CRUD operations
    """

    @abstractmethod
    def get_one(self, pk: t.Any, **kwargs: t.Any) -> t.Any:
        pass

    @abstractmethod
    def get_all(self, **kwargs: t.Any) -> t.Union[QuerySet, t.List[t.Any]]:
        pass

    @abstractmethod
    def create(self, schema: PydanticModel, **kwargs: t.Any) -> t.Any:
        pass

    @abstractmethod
    def update(
        self, instance: DjangoModel, schema: PydanticModel, **kwargs: t.Any
    ) -> t.Any:
        pass

    @abstractmethod
    def patch(
        self, instance: DjangoModel, schema: PydanticModel, **kwargs: t.Any
    ) -> t.Any:
        pass

    @abstractmethod
    def delete(self, instance: DjangoModel, **kwargs: t.Any) -> t.Any:
        pass



from django.conf import settings
from django.http import HttpRequest

from ninja_extra.security.api_key import AsyncAPIKeyCookie

__all__ = ["AsyncSessionAuth"]

from typing import Any, Optional


class AsyncSessionAuth(AsyncAPIKeyCookie):
    """Reusing Django session authentication"""

    param_name: str = settings.SESSION_COOKIE_NAME

    async def authenticate(
        self, request: HttpRequest, key: Optional[str]
    ) -> Optional[Any]:
        if request.user.is_authenticated:
            return request.user
        return None



import dataclasses
from typing import Any, Dict, Generic, List, Optional, Type, TypeVar, Union

from ninja import Schema
from ninja.constants import NOT_SET, NOT_SET_TYPE
from ninja.throttling import BaseThrottle
from pydantic import BeforeValidator, TypeAdapter, field_validator
from pydantic.networks import HttpUrl
from typing_extensions import Annotated

T = TypeVar("T")

Url = Annotated[
    str, BeforeValidator(lambda value: str(TypeAdapter(HttpUrl).validate_python(value)))
]


class BasePaginatedResponseSchema(Schema):
    count: int
    next: Optional[Url]
    previous: Optional[Url]
    results: List[Any]


class BaseNinjaResponseSchema(Schema):
    count: int
    items: List[Any]


class PaginatedResponseSchema(BasePaginatedResponseSchema, Generic[T]):
    results: List[T]


# Pydantic GenericModels has not way of identifying the _orig
# __generic_model__ is more like a fix for that
# PaginatedResponseSchema.__generic_model__ = (  # type:ignore[attr-defined]
#     PaginatedResponseSchema
# )


class NinjaPaginationResponseSchema(BaseNinjaResponseSchema, Generic[T]):
    items: List[T]

    @field_validator("items", mode="before")
    def validate_items(cls, value: Any) -> Any:
        if value is not None and not isinstance(value, list):
            value = list(value)
        return value


# NinjaPaginationResponseSchema.__generic_model__ = (  # type:ignore[attr-defined]
#     NinjaPaginationResponseSchema
# )


@dataclasses.dataclass
class RouteParameter:
    path: str
    methods: List[str]
    openapi_extra: Optional[Dict[str, Any]]
    auth: Optional[Union[Type, Any]] = None
    throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET
    response: Optional[Union[Type, Any]] = None
    operation_id: Optional[str] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    tags: Optional[List[str]] = None
    deprecated: Optional[bool] = None
    by_alias: bool = False
    exclude_unset: bool = False
    exclude_defaults: bool = False
    exclude_none: bool = False
    url_name: Optional[str] = None
    include_in_schema: bool = True

    def dict(self) -> dict:
        return dataclasses.asdict(self)


def __getattr__(name: str) -> Any:  # pragma: no cover
    if name in [
        "IdSchema",
        "OkSchema",
        "DetailSchema",
    ]:
        raise RuntimeError(f"'{name}' is no longer available")



import typing as t

from ninja.orm.fields import TYPES

from ninja_extra.constants import ROUTE_FUNCTION

from .endpoints import ModelAsyncEndpointFactory, ModelEndpointFactory
from .schemas import ModelConfig

if t.TYPE_CHECKING:
    from ..base import APIController, ModelControllerBase


class ModelControllerBuilder:
    """
    Model Controller Builder that controls model controller setup.
    """

    def __init__(
        self,
        base_cls: t.Type["ModelControllerBase"],
        api_controller_instance: "APIController",
    ) -> None:
        assert base_cls.model_config
        self._config: ModelConfig = base_cls.model_config
        self._base_cls = base_cls
        self._api_controller_instance = api_controller_instance
        model_pk = getattr(
            self._config.model._meta.pk,
            "name",
            self._config.model._meta.pk.attname,  # type:ignore[union-attr]
        )
        internal_type = (
            self._config.model._meta.pk.get_internal_type()  # type:ignore[union-attr]
        )
        self._pk_type: t.Type = TYPES.get(internal_type, str)
        self._model_pk_name = model_pk
        self._model_name = self._config.model.__name__.replace("Model", "")

        self._retrieve_schema = self._config.retrieve_schema
        self._create_schema = self._config.create_schema
        self._update_schema = self._config.update_schema
        self._patch_schema = self._config.patch_schema

        self._route_factory: ModelEndpointFactory = (
            ModelAsyncEndpointFactory()
            if base_cls.model_config.async_routes
            else ModelEndpointFactory()
        )

    def _add_to_controller(self, func: t.Callable) -> None:
        route_function = getattr(func, ROUTE_FUNCTION)
        route_function.api_controller = self._api_controller_instance
        self._api_controller_instance.add_controller_route_function(route_function)

    def _register_create_endpoint(self) -> None:
        kw = {
            "url_name": f"{self._model_name.lower()}-create",
            "description": f"Create {self._model_name} item",
            "summary": "Create an item",
        }
        kw.update(self._config.create_route_info)
        create_item = self._route_factory.create(
            schema_in=self._create_schema,  # type:ignore[arg-type]
            schema_out=self._retrieve_schema,  # type:ignore[arg-type]
            **kw,  # type:ignore[arg-type]
        )

        self._add_to_controller(create_item)

    def _register_update_endpoint(self) -> None:
        _path = "/{%s:%s}" % (
            self._pk_type.__name__,
            self._model_pk_name,
        )
        kw = {
            "url_name": f"{self._model_name.lower()}-put",
            "description": f"""Update {self._model_name} item by {self._model_pk_name}""",
            "summary": "Update an item",
        }
        kw.update(self._config.update_route_info)

        update_item = self._route_factory.update(
            path=_path,
            lookup_param=self._model_pk_name,
            schema_in=self._update_schema,  # type:ignore[arg-type]
            schema_out=self._retrieve_schema,  # type:ignore[arg-type]
            **kw,  # type:ignore[arg-type]
        )

        self._add_to_controller(update_item)

    def _register_patch_endpoint(self) -> None:
        _pk_type = self._pk_type
        _path = "/{%s:%s}" % (
            _pk_type.__name__,
            self._model_pk_name,
        )

        kw = {
            "url_name": f"{self._model_name.lower()}-patch",
            "description": f"""Patch {self._model_name} item by {self._model_pk_name}""",
            "summary": "Patch an item",
        }
        kw.update(self._config.patch_route_info)

        patch_item = self._route_factory.patch(
            path=_path,
            lookup_param=self._model_pk_name,
            schema_out=self._retrieve_schema,  # type:ignore[arg-type]
            schema_in=self._patch_schema,  # type:ignore[arg-type]
            **kw,  # type:ignore[arg-type]
        )

        self._add_to_controller(patch_item)

    def _register_find_one_endpoint(self) -> None:
        _path = "/{%s:%s}" % (
            self._pk_type.__name__,
            self._model_pk_name,
        )
        kw = {
            "url_name": f"{self._model_name.lower()}-get-item",
            "description": f"""Get {self._model_name} item by {self._model_pk_name}""",
            "summary": "Get a specific item",
        }
        kw.update(self._config.find_one_route_info)

        get_item = self._route_factory.find_one(
            path=_path,
            lookup_param=self._model_pk_name,
            schema_out=self._retrieve_schema,  # type:ignore[arg-type]
            **kw,  # type:ignore[arg-type]
        )

        self._add_to_controller(get_item)

    def _register_list_endpoint(self) -> None:
        kw = {
            "description": f"List {self._model_name} model items",
            "url_name": f"{self._model_name.lower()}-list",
            "summary": "List Items",
        }
        kw.update(self._config.list_route_info)
        paginate_kwargs: t.Dict[str, t.Any] = {
            "pagination_class": None,
            "pagination_response_schema": None,
        }
        if self._config.pagination:
            paginate_kwargs.update(
                pagination_class=self._config.pagination.klass,
                pagination_response_schema=self._config.pagination.pagination_schema,
            )
            if self._config.pagination.paginator_kwargs:  # pragma: no cover
                paginate_kwargs.update(self._config.pagination.paginator_kwargs)

        list_items = self._route_factory.list(
            path="/",
            schema_out=self._retrieve_schema,  # type:ignore[arg-type]
            **kw,  # type:ignore[arg-type]
            **paginate_kwargs,
        )

        self._add_to_controller(list_items)

    def _register_delete_endpoint(self) -> None:
        _path = "/{%s:%s}" % (
            self._pk_type.__name__,
            self._model_pk_name,
        )
        kw = {
            "url_name": f"{self._model_name.lower()}-delete",
            "description": f"""Delete {self._model_name} item""",
            "summary": "Delete an item",
        }
        kw.update(self._config.delete_route_info)

        delete_item = self._route_factory.delete(
            path=_path,
            lookup_param=self._model_pk_name,
            **kw,  # type:ignore[arg-type]
        )

        self._add_to_controller(delete_item)

    def register_model_routes(self) -> None:
        for action in self._config.allowed_routes:
            action_registration = getattr(self, f"_register_{action}_endpoint", None)

            if not action_registration:  # pragma: no cover
                raise Exception(
                    f"Route `{action}` action in `class[{self._base_cls.__name__}]` "
                    f"is not recognized as ModelController action"
                )
            action_registration()



import contextvars
import typing as t

if t.TYPE_CHECKING:  # pragma: no cover
    from ninja_extra.controllers.route.context import RouteContext


POST = "POST"
PUT = "PUT"
PATCH = "PATCH"
DELETE = "DELETE"
GET = "GET"
HEAD = "HEAD"
OPTIONS = "OPTIONS"
TRACE = "TRACE"
ROUTE_METHODS = [POST, PUT, PATCH, DELETE, GET, HEAD, OPTIONS, TRACE]
THROTTLED_FUNCTION = "__throttled_endpoint__"
THROTTLED_OBJECTS = "__throttled_objects__"
ROUTE_FUNCTION = "__route_function__"

ROUTE_CONTEXT_VAR: contextvars.ContextVar[t.Optional["RouteContext"]] = (
    contextvars.ContextVar("ROUTE_CONTEXT_VAR")
)
ROUTE_CONTEXT_VAR.set(None)



import datetime
import functools
import re
import typing as t
import uuid
from urllib.parse import parse_qs, urlparse

from ninja import Query, Schema
from ninja.params import Path
from pydantic import UUID4, create_model

from ninja_extra.shortcuts import add_ninja_contribute_args

# Match parameters in URL paths, eg. '{param}', and '{int:param}'
PARAM_REGEX = re.compile("{([a-zA-Z_][a-zA-Z0-9_]*:)?([a-zA-Z_][a-zA-Z0-9_]*)}")


class PathCompiledResult(t.NamedTuple):
    param_convertors: t.Dict[str, t.Any]
    query_parameters: t.Dict[str, t.Any]

    def has_any_parameter(self) -> bool:
        return len(self.param_convertors) > 0 or len(self.query_parameters) > 0


STRING_TYPES: t.Dict[str, t.Type] = {
    "int": int,
    "path": str,
    "slug": str,
    "str": str,
    "uuid": UUID4,
    "datetime": datetime.datetime,
    "date": datetime.date,
}


def compile_path(path: str) -> PathCompiledResult:
    """
    Given a path string, like: "/{str:username}"

    regex:      "/(?P<username>[^/]+)"
    format:     "/{username}"
    convertors: {str:"username"}
    """
    duplicated_params = set()
    param_convertors = {}

    parsed_url = urlparse(path)
    query_parameters = parse_qs(parsed_url.query)

    for match in PARAM_REGEX.finditer(path):
        convertor_type, param_name = match.groups("str")
        convertor = convertor_type.rstrip(":")

        if param_name in param_convertors:
            duplicated_params.add(param_name)

        param_convertors[param_name] = STRING_TYPES[convertor]

    if duplicated_params:
        names = ", ".join(sorted(duplicated_params))
        ending = "s" if len(duplicated_params) > 1 else ""
        raise ValueError(f"Duplicated param name{ending} {names} at path {path}")

    return PathCompiledResult(param_convertors, query_parameters)


class PathResolverOperation:
    """
    Create a decorator to parse `path` parameters and resolve them during endpoint.
    For example:
    path=/{int:id}/tags/{post_id}?query=int&query1=int
    this will create two path parameters `id` and `post_id` and two query parameters `query` and `query1`
    for the decorated endpoint
    """

    def __init__(self, path: str, func: t.Callable) -> None:
        self.compiled_path = compile_path(path)
        self._view_func = func

        if self.compiled_path.has_any_parameter():
            _ninja_contribute_args: t.List[t.Tuple] = getattr(
                func, "_ninja_contribute_args", []
            )

            unique_key = str(uuid.uuid4().hex)[:5]

            self.path_construct_name = f"PathModel{unique_key}"
            self.query_construct_name = f"QueryModel{unique_key}"

            path_fields = dict(self.get_path_fields())
            query_fields = dict(self.get_query_fields())

            if path_fields:
                dynamic_path_model = create_model(
                    self.path_construct_name, __base__=Schema, **path_fields
                )
                add_ninja_contribute_args(
                    func,
                    (self.path_construct_name, dynamic_path_model, Path(...)),
                )

            if query_fields:
                dynamic_query_model = create_model(
                    self.query_construct_name, __base__=Schema, **query_fields
                )

                add_ninja_contribute_args(
                    func,
                    (self.query_construct_name, dynamic_query_model, Query(...)),
                )
            self.as_view = functools.wraps(func)(self.get_view_function())
        else:
            self.as_view = func  # type:ignore[assignment]

    def get_path_fields(self) -> t.Generator:
        for path_name, path_type in self.compiled_path.param_convertors.items():
            yield path_name, (path_type, ...)

    def get_query_fields(self) -> t.Generator:
        for path_name, path_type in self.compiled_path.query_parameters.items():
            path_type.append("str")
            yield path_name, (STRING_TYPES[path_type[0]], ...)

    def get_view_function(self) -> t.Callable:
        def as_view(*args: t.Any, **kwargs: t.Any) -> t.Any:
            func_kwargs = dict(kwargs)

            if self.path_construct_name in func_kwargs:
                dynamic_model_path_instance = func_kwargs.pop(self.path_construct_name)
                func_kwargs.update(dynamic_model_path_instance.dict())

            if self.query_construct_name in func_kwargs:
                dynamic_model_query_instance = func_kwargs.pop(
                    self.query_construct_name
                )
                func_kwargs.update(dynamic_model_query_instance.dict())

            return self._view_func(*args, **func_kwargs)

        return as_view


class AsyncPathResolverOperation(PathResolverOperation):
    def get_view_function(self) -> t.Callable:
        async def as_view(*args: t.Any, **kwargs: t.Any) -> t.Any:
            func_kwargs = dict(kwargs)

            if self.path_construct_name in func_kwargs:
                dynamic_model_path_instance = func_kwargs.pop(self.path_construct_name)
                func_kwargs.update(dynamic_model_path_instance.dict())

            if self.query_construct_name in func_kwargs:
                dynamic_model_query_instance = func_kwargs.pop(
                    self.query_construct_name
                )
                func_kwargs.update(dynamic_model_query_instance.dict())

            return await self._view_func(*args, **func_kwargs)

        return as_view



from ninja.security import (
    APIKeyCookie,
    APIKeyHeader,
    APIKeyQuery,
    HttpBasicAuth,
    HttpBearer,
    SessionAuth,
    django_auth,
)

from .api_key import (
    AsyncAPIKeyBase,
    AsyncAPIKeyCookie,
    AsyncAPIKeyHeader,
    AsyncAPIKeyQuery,
)
from .http import AsyncHttpBasicAuth, AsyncHttpBearer
from .session import AsyncSessionAuth

__all__ = [
    "APIKeyCookie",
    "APIKeyHeader",
    "APIKeyQuery",
    "HttpBasicAuth",
    "HttpBearer",
    "SessionAuth",
    "django_auth",
    "async_django_auth",
    "AsyncHttpBasicAuth",
    "AsyncHttpBasicAuth",
    "AsyncHttpBearer",
    "AsyncAPIKeyCookie",
    "AsyncAPIKeyBase",
    "AsyncAPIKeyHeader",
    "AsyncAPIKeyQuery",
]

async_django_auth = AsyncSessionAuth()



from json import dumps as json_dumps
from typing import Any, Callable, Dict, Optional, Type, Union
from unittest.mock import Mock
from urllib.parse import urlencode

from ninja import NinjaAPI, Router
from ninja.responses import NinjaJSONEncoder
from ninja.testing.client import NinjaClientBase, NinjaResponse

from ninja_extra import ControllerBase, NinjaExtraAPI


class NinjaExtraClientBase(NinjaClientBase):
    def __init__(
        self, router_or_app: Union[NinjaAPI, Router, Type[ControllerBase]], **kw: Any
    ) -> None:
        if hasattr(router_or_app, "get_api_controller"):
            api = NinjaExtraAPI(**kw)
            controller_ninja_api_controller = router_or_app.get_api_controller()
            assert controller_ninja_api_controller
            controller_ninja_api_controller.set_api_instance(api)
            self._urls_cache = list(controller_ninja_api_controller.urls_paths(""))
            router_or_app = api
        super(NinjaExtraClientBase, self).__init__(router_or_app)

    def request(
        self,
        method: str,
        path: str,
        data: Optional[Dict] = None,
        json: Any = None,
        **request_params: Any,
    ) -> "NinjaResponse":
        if json is not None:
            request_params["body"] = json_dumps(json, cls=NinjaJSONEncoder)
        if "query" in request_params and isinstance(request_params["query"], dict):
            query = request_params.pop("query")
            url_encode = urlencode(query)
            path = f"{path}?{url_encode}"
        func, request, kwargs = self._resolve(
            method, path, {} if data is None else data, request_params
        )
        return self._call(func, request, kwargs)  # type: ignore


class TestClient(NinjaExtraClientBase):
    def _call(self, func: Callable, request: Mock, kwargs: Dict) -> "NinjaResponse":
        return NinjaResponse(func(request, **kwargs))


class TestAsyncClient(NinjaExtraClientBase):
    async def _call(self, func: Callable, request: Mock, kwargs: Dict) -> NinjaResponse:
        return NinjaResponse(await func(request, **kwargs))



from typing import (
    Any,
    Optional,
    Type,
    Union,
    no_type_check,
)

from ninja import Schema

from .. import status


class ControllerResponse:
    status_code: int = status.HTTP_204_NO_CONTENT
    _schema: Optional[Any]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        raise RuntimeError("Controller Response are no longer supported.")

    @classmethod
    def get_schema(cls) -> Union[Schema, Type[Schema], Any]:  # pragma: no cover
        raise NotImplementedError

    def convert_to_schema(self) -> Any:  # pragma: no cover
        raise NotImplementedError

    @no_type_check
    def __class_getitem__(cls: Type["ControllerResponse"], item: Any) -> Any:
        raise RuntimeError("Controller Response are no longer supported.")


Detail = ControllerResponse
Ok = ControllerResponse
Id = ControllerResponse



import traceback
import typing as t

from asgiref.sync import sync_to_async
from django.db.models import Model, QuerySet
from pydantic import BaseModel as PydanticModel

from ninja_extra.exceptions import NotFound
from ninja_extra.shortcuts import get_object_or_exception

from .interfaces import AsyncModelServiceBase, ModelServiceBase


class ModelService(ModelServiceBase, AsyncModelServiceBase):
    """
    Model Service for Model Controller model CRUD operations with a simple logic for simple models.

    Its advised you override this class if you have a complex model.
    """

    def __init__(self, model: t.Type[Model]) -> None:
        self.model = model

    def get_one(self, pk: t.Any, **kwargs: t.Any) -> t.Any:
        obj = get_object_or_exception(
            klass=self.model, error_message=None, exception=NotFound, pk=pk
        )
        return obj

    async def get_one_async(self, pk: t.Any, **kwargs: t.Any) -> t.Any:
        return await sync_to_async(self.get_one, thread_sensitive=True)(pk, **kwargs)

    def get_all(self, **kwargs: t.Any) -> t.Union[QuerySet, t.List[t.Any]]:
        return self.model.objects.all()

    async def get_all_async(self, **kwargs: t.Any) -> t.Union[QuerySet, t.List[t.Any]]:
        return await sync_to_async(self.get_all, thread_sensitive=True)(**kwargs)

    def create(self, schema: PydanticModel, **kwargs: t.Any) -> t.Any:
        data = schema.model_dump(by_alias=True)
        data.update(kwargs)

        try:
            instance = self.model._default_manager.create(**data)
            return instance
        except TypeError as tex:  # pragma: no cover
            tb = traceback.format_exc()
            msg = (
                "Got a `TypeError` when calling `%s.%s.create()`. "
                "This may be because you have a writable field on the "
                "serializer class that is not a valid argument to "
                "`%s.%s.create()`. You may need to make the field "
                "read-only, or override the %s.create() method to handle "
                "this correctly.\nOriginal exception was:\n %s"
                % (
                    self.model.__name__,
                    self.model._default_manager.name,
                    self.model.__name__,
                    self.model._default_manager.name,
                    self.__class__.__name__,
                    tb,
                )
            )
            raise TypeError(msg) from tex

    async def create_async(self, schema: PydanticModel, **kwargs: t.Any) -> t.Any:
        return await sync_to_async(self.create, thread_sensitive=True)(schema, **kwargs)

    def update(self, instance: Model, schema: PydanticModel, **kwargs: t.Any) -> t.Any:
        data = schema.model_dump(exclude_none=True)
        data.update(kwargs)
        for attr, value in data.items():
            setattr(instance, attr, value)
        instance.save()
        return instance

    async def update_async(
        self, instance: Model, schema: PydanticModel, **kwargs: t.Any
    ) -> t.Any:
        return await sync_to_async(self.update, thread_sensitive=True)(
            instance, schema, **kwargs
        )

    def patch(self, instance: Model, schema: PydanticModel, **kwargs: t.Any) -> t.Any:
        return self.update(instance=instance, schema=schema, **kwargs)

    async def patch_async(
        self, instance: Model, schema: PydanticModel, **kwargs: t.Any
    ) -> t.Any:
        return await self.update_async(instance=instance, schema=schema, **kwargs)

    def delete(self, instance: Model, **kwargs: t.Any) -> t.Any:
        instance.delete()

    async def delete_async(self, instance: Model, **kwargs: t.Any) -> t.Any:
        return await sync_to_async(self.delete, thread_sensitive=True)(instance)



import logging
from collections import OrderedDict
from typing import (
    Any,
    Optional,
    Type,
    Union,
)

from django.core.paginator import InvalidPage, Page, Paginator
from django.db.models import QuerySet
from django.http import HttpRequest
from ninja import Schema
from ninja.pagination import PaginationBase
from ninja.types import DictStrAny
from pydantic import Field

from ninja_extra.conf import settings
from ninja_extra.exceptions import NotFound
from ninja_extra.schemas import PaginatedResponseSchema
from ninja_extra.urls import remove_query_param, replace_query_param

logger = logging.getLogger()


class PageNumberPaginationExtra(PaginationBase):
    class Input(Schema):
        page: int = Field(1, gt=0)
        page_size: int = Field(100, lt=200)

    page_query_param = "page"
    page_size_query_param = "page_size"

    max_page_size = 200
    paginator_class = Paginator

    def __init__(
        self,
        page_size: int = settings.PAGINATION_PER_PAGE,
        max_page_size: Optional[int] = None,
        pass_parameter: Optional[str] = None,
    ) -> None:
        super().__init__(pass_parameter=pass_parameter)
        self.page_size = page_size
        self.max_page_size = max_page_size or 200
        self.Input = self.create_input()  # type:ignore

    def create_input(self) -> Type[Input]:
        class DynamicInput(PageNumberPaginationExtra.Input):
            page: int = Field(1, gt=0)
            page_size: int = Field(self.page_size, lt=self.max_page_size)

        return DynamicInput

    def paginate_queryset(
        self,
        queryset: QuerySet,
        pagination: Input,
        request: Optional[HttpRequest] = None,
        **params: DictStrAny,
    ) -> Any:
        assert request, "request is required"
        current_page_number = pagination.page
        paginator = self.paginator_class(queryset, pagination.page_size)
        try:
            url = request.build_absolute_uri()
            page: Page = paginator.page(current_page_number)
            return self.get_paginated_response(base_url=url, page=page)
        except InvalidPage as exc:  # pragma: no cover
            msg = "Invalid page. {page_number} {message}".format(
                page_number=current_page_number, message=str(exc)
            )
            raise NotFound(msg) from exc

    def get_paginated_response(self, *, base_url: str, page: Page) -> DictStrAny:
        return OrderedDict(
            [
                ("count", page.paginator.count),
                ("next", self.get_next_link(base_url, page=page)),
                ("previous", self.get_previous_link(base_url, page=page)),
                ("results", list(page)),
            ]
        )

    @classmethod
    def get_response_schema(
        cls, response_schema: Union[Type[Schema], Type[Any]]
    ) -> Any:
        return PaginatedResponseSchema[response_schema]  # type: ignore[valid-type]

    def get_next_link(self, url: str, page: Page) -> Optional[str]:
        if not page.has_next():
            return None
        page_number = page.next_page_number()
        return replace_query_param(url, self.page_query_param, page_number)

    def get_previous_link(self, url: str, page: Page) -> Optional[str]:
        if not page.has_previous():
            return None
        page_number = page.previous_page_number()
        if page_number == 1:
            return remove_query_param(url, self.page_query_param)
        return replace_query_param(url, self.page_query_param, page_number)
